import numpy as np
from numba import jit, prange
from numba import NumbaDeprecationWarning, NumbaPendingDeprecationWarning, NumbaWarning
import warnings

warnings.simplefilter('ignore', category=NumbaDeprecationWarning)
warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)
warnings.simplefilter('ignore', category=NumbaWarning)


@jit("float64[:](int8, int8)",parallel=True)
def Levy(n, T):
    y = np.zeros(n)
    D = np.random.exponential(2 * T)
    d = int(D)
    for i in prange(n):
        y[i] =np.tan(np.random.uniform(0, np.pi, d)).sum()
    return y-0.5



import numpy as np


class InitOpt_Delta_z:
    @staticmethod
    def your():
        pass
    @staticmethod
    def goa(t, z_interval, T):
        z = z_interval[1] - t * ((z_interval[1] - z_interval[0]) / T)
        return z

    @staticmethod
    def ba1(t,z1_0, w1):
        z1_0 = z1_0
        new_z1 = z1_0 * (1 - np.exp(-w1 * t))
        return new_z1

    @staticmethod
    def ba2(old_z2, w2):
        new_z2 = w2 * old_z2
        return new_z2

    @staticmethod
    def boa(t, old_z, budget):
        # sensory_modality_update in original paper
        if t == 0:
            new_z = old_z
            return new_z
        else:
            new_z = old_z + 0.025 / (old_z*budget)
            return new_z

    @staticmethod
    def mbo(t, w4):
        t = t + 1
        return w4 / (t * t)

    @staticmethod
    def mfo1(t,budget):
        return -1 - t / budget

    @staticmethod
    def mfo2(t, M, budget):
        return round(M - t * ((M - 1) /budget))



from sklearn.metrics import pairwise_distances
from .LevyFlight import Levy
import numpy as np
from numba import jit, prange

class Opt_X:
    @staticmethod
    def your():
        pass

    @staticmethod
    def pso(old_X, new_Y, lb_x, ub_x):
        new_X = old_X.copy()
        for i, y in enumerate(new_Y):
            x = old_X[i] + y  # update equation.
            new_X[i] = np.clip(x, lb_x, ub_x)
        return new_X

    @staticmethod
    def mfo(old_X, new_Y, z1, z2, w, lb_x, ub_x):
        new_X = old_X.copy()
        for i, y in enumerate(new_Y):
            A = abs(y - old_X[i])
            B = (z1 - 1)*np.random.rand() + 1
            C = A * np.exp(w * B) * np.cos(B * 2 * np.pi)
            if i <= z2:
                x = C + y
            else:
                x = C + new_Y[z2]
            new_X[i] = np.clip(x, lb_x, ub_x)
        return new_X


    @staticmethod
    def ba(old_X, new_Y, old_x_g, old_z1, old_z2, w3, lb_x, ub_x):
        new_X = old_X.copy()
        n = old_X.shape[1]
        for i, y in enumerate(new_Y):
            x = old_X[i] + y
            if np.random.rand() < old_z1:
                x = old_x_g + w3 * np.random.randn(n) * old_z2
            new_X[i] = np.clip(x, lb_x, ub_x)
        return new_X

    @staticmethod
    def csa(new_Y, old_X, w1, w2, lb_x, ub_x):
        new_X = old_X
        M, n = old_X.shape[0], old_X.shape[1]
        for i in range(M):
            if np.random.rand() > w1:
                new_X[i] = np.clip(old_X[i] + np.random.rand() * w2 * (new_Y[np.random.randint(M)] - old_X[i]), lb_x, ub_x)
            else:
                new_X[i] = np.random.uniform(lb_x, ub_x, n)
        return new_X

    @staticmethod
    @jit(parallel=True)
    def goa(old_X, old_x_g, z, lb_x, ub_x, w1, w2):
        M, n = old_X.shape[0], old_X.shape[1]
        new_X =old_X.copy()
        Dist = pairwise_distances(old_X, metric='euclidean')
        for i in prange(M):
            Delta_X = np.zeros(n)
            for j in range(M):
                if i != j:
                    # js are neighbors
                    D = Dist[i, j]
                    norm_D = 2 + np.remainder(D, 2)
                    delta_x = z * ((ub_x - lb_x) / 2) * \
                              (w1 * np.exp(-norm_D/w2)-np.exp(-norm_D)) * \
                              ((old_X[i] - old_X[j])/(D+2.2204e-16))
                    Delta_X = Delta_X + delta_x
            new_X[i] = np.clip(z * Delta_X + old_x_g, lb_x, ub_x)
        return new_X

    @staticmethod
    @jit(parallel=True)
    def mbo(old_X, old_X_Fit, x_g, z, w1, w2, w3, budget):
        M, n = old_X.shape[0], old_X.shape[1]
        I_s = np.argsort(old_X_Fit)
        sorted_X = old_X[I_s]
        n_strongs = int(np.ceil(w1 * M))
        n_weaks = M - n_strongs
        strong_pop = sorted_X[0:n_strongs]
        weak_pop = sorted_X[n_strongs:]
        # strong
        new_strongs = strong_pop.copy()
        tfs_strong = np.random.rand(n_strongs, n) * w2 <= w1
        for i in prange(n_strongs):
            for d in prange(n):
                if tfs_strong[i, d]:
                    new_strongs[i, d] = strong_pop[np.random.randint(n_strongs), d]
                else:
                    new_strongs[i, d] = weak_pop[np.random.randint(n_weaks), d]

        # weak
        new_weaks = weak_pop.copy()
        tfs_weak = np.random.rand(n_weaks, n) <= w1
        for i in prange(n_weaks):
            for d in prange(n):
                if tfs_weak[i, d]:
                    new_weaks[i, d] = x_g[d]
                else:
                    y = Levy(n=n, T=budget)
                    new_weaks[i, d] = weak_pop[np.random.randint(n_weaks), d]
                    if np.random.rand() > w3:
                        new_weaks[i, d] = new_weaks[i, d] + z * y[d]

        temp_X = np.concatenate((new_strongs, new_weaks))
        return temp_X


    @staticmethod
    def boa(old_X, old_X_Fit, x_g, z, w1, w2,lb_x, ub_x):
        M = len(old_X)
        new_X = old_X.copy()
        for i in range(M):
            if np.random.rand()>w2:
                temp = np.random.rand() ** 2 * x_g - old_X[i]
                x = old_X[i] + temp * z * old_X_Fit[i] ** w1
            else:
                jk = np.random.choice(M, 2, replace=False) # two indices of two neighbors
                temp = np.random.rand()**2*old_X[jk[0]]-old_X[jk[1]]
                x = old_X[i] + temp * z * old_X_Fit[i] ** w1
            new_X[i] = np.clip(x, lb_x, ub_x)
        return new_X





import numpy as np

class Opt_Delta_X:
    @staticmethod
    def your():
        pass

    @staticmethod
    def Personal_best(new_X, new_X_Fit, old_X_p, old_X_p_Fit):
        new_X_p = old_X_p
        new_X_p_Fit = old_X_p_Fit
        for i, x in enumerate(new_X):
            if new_X_Fit[i] < old_X_p_Fit[i]:
                new_X_p[i] = x
                new_X_p_Fit[i] = new_X_Fit[i]
        return new_X_p, new_X_p_Fit

    @staticmethod
    def Global_best(new_X, new_X_Fit, old_x_g, old_x_g_fit):
        xs = np.append([old_x_g],new_X, axis=0)
        fits = np.append([old_x_g_fit], new_X_Fit, axis=0)
        best_index = np.argmin(fits)
        x_g_fit = fits[best_index]
        x_g = xs[best_index]
        return x_g, x_g_fit


from UNIOA.NatureOpt import NatureOpt


# -------------------------------------------------------------------------------------------------
# This class implements BOA-Optimizer in the new structure.
# -------------------------------------------------------------------------------------------------

class BOA_Opt(NatureOpt):
    def __init__(self, func ,hyperparams_set, budget_factor=1e4):
        super().__init__(func, budget_factor)
        self.M = hyperparams_set.get('popsize', 50)
        self.w1 = hyperparams_set.get('w1', 0.1)#=power exponent
        self.w2 = hyperparams_set.get('w2', 0.8)#=prob
        self.z_0 = hyperparams_set.get('z_0', 0.01)# = sensory modality


    def __call__(self):
        t = 0
        X = self.Init_X.Init_X(M=self.M, n=self.n, lb_x=self.lb_x, ub_x=self.ub_x)
        X_Fit = self.Evaluate_X(X=X)
        x_g, x_g_fit = self.Init_Delta_X.Global_best(new_X=X, new_X_Fit=X_Fit)
        z = self.InitOpt_Delta_z.boa(t=t, old_z=self.z_0, budget=self.budget)

        while not self.stop:
            # Opt temp_x(t+1)
            temp_X = self.Opt_X.boa(old_X=X, old_X_Fit=X_Fit, x_g=x_g, z=z, w1=self.w1, w2=self.w2, lb_x=self.lb_x, ub_x=self.ub_x)
            temp_X_Fit = self.Evaluate_X(X=temp_X)
            # Selection
            new_X, new_X_Fit = self.Selection.improve_type(temp_X=temp_X, temp_X_Fit=temp_X_Fit, old_X=X, old_X_Fit=X_Fit)
            # ----------------------------
            t = t + 1
            new_x_g, new_x_g_fit = self.Opt_Delta_X.Global_best(new_X=new_X, new_X_Fit=new_X_Fit, old_x_g=x_g, old_x_g_fit=x_g_fit)
            new_z = self.InitOpt_Delta_z.boa(t=t, old_z=z, budget=self.budget)
            #######################################
            X = new_X
            X_Fit = new_X_Fit
            x_g, x_g_fit = new_x_g, new_x_g_fit
            z = new_z


import numpy as np

class Init_Delta_X:
    @staticmethod
    def your():
        pass

    @staticmethod
    def Personal_best(new_X, new_X_Fit):
        new_X_p = new_X
        new_X_p_Fit = new_X_Fit
        return new_X_p, new_X_p_Fit

    @staticmethod
    def Global_best(new_X, new_X_Fit):
        best_index = np.argmin(new_X_Fit)
        x_g_fit = new_X_Fit[best_index]
        x_g = new_X[best_index]
        return x_g, x_g_fit




from UNIOA.NatureOpt import NatureOpt

# -------------------------------------------------------------------------------------------------
# MothFlame-Optimizer in the new structure.
# -------------------------------------------------------------------------------------------------

class MFO_Opt(NatureOpt):
    def __init__(self, func ,hyperparams_set, budget_factor=1e4):
        super().__init__(func, budget_factor)
        self.M = hyperparams_set.get('popsize', 30)
        self.w = hyperparams_set.get('w', 1)

    def __call__(self):
        t = 0
        X = self.Init_X.Init_X(M=self.M, n=self.n, lb_x=self.lb_x, ub_x=self.ub_x)
        X_Fit = self.Evaluate_X(X=X)
        Y = self.Init_Delta_Y.x_type(X=X)
        z1 = self.InitOpt_Delta_z.mfo1(t=t, budget=self.budget)
        z2 = self.InitOpt_Delta_z.mfo2(t=t, M=self.M, budget=self.budget)

        # optimize process
        while not self.stop:
            # AOptimize y(t+1)
            new_Y = self.Opt_Delta_Y.mfo(t=t, old_Y=Y, fitness_function=self.fitness_function, old_X=X, old_X_Fit=X_Fit)
            # OOptimize temp_X(t+1)
            temp_X = self.Opt_X.mfo(old_X=X, new_Y=new_Y, z1=z1, z2=z2, w=self.w, lb_x=self.lb_x, ub_x=self.ub_x)
            # Evaluate
            temp_X_Fit = self.Evaluate_X(X=temp_X)
            # Selection
            new_X, new_X_Fit = self.Selection.same_type(temp_X=temp_X, temp_X_Fit=temp_X_Fit)

            # ----------------------------
            t = t + 1
            new_z1 = self.InitOpt_Delta_z.mfo1(t=t, budget=self.budget)
            new_z2 = self.InitOpt_Delta_z.mfo2(t=t, M=self.M, budget=self.budget)
            ##################################
            X = new_X
            X_Fit= new_X_Fit
            Y = new_Y
            z1 = new_z1
            z2 = new_z2


import time
import ioh
from datetime import datetime
from .algs import *
import sys, os
from pydoc import locate


def optimizer_running(problems, instances, dimensions, num_runs, paras_set, optimizer_name):
    t = 0
    optimizer_name_temp = optimizer_name
    UNIOA_algs = ['BA_Opt', 'CSA_Opt', 'MFO_Opt', 'PSO_Opt', 'GOA_Opt', 'MBO_Opt', 'BOA_Opt']
    if optimizer_name not in UNIOA_algs:
        t = 1
        file_name = os.path.basename(sys.argv[0])[:-2]
        your_optimizer_name = locate(file_name + optimizer_name)


    folder_name = optimizer_name_temp + '_folder'
    data_name = optimizer_name_temp
    st = time.time()
    print('----- ' + optimizer_name_temp + ' is optmizting your problem -----', flush=True)
    logger = ioh.logger.Analyzer(folder_name='DataFiles/'+folder_name+'/'+data_name, algorithm_name=optimizer_name_temp)
    for p_id in problems:
        print('P-id={}'.format(p_id), end=":", flush=True)
        for d in dimensions:
            print('d-{}'.format(d), end=":", flush=True)
            for i_id in instances:
                print('{}'.format(i_id), end="", flush=True)
                func = ioh.get_problem(fid=p_id, dim=d, iid=i_id)
                # func.attach_logger(logger)
                print('', end='(', flush=True)
                for rep in range(num_runs):
                    func.attach_logger(logger)
                    print('{}'.format(rep), end="", flush=True)
                    if t == 1 :
                        opt = your_optimizer_name(func, paras_set)
                    else:
                        opt = eval(optimizer_name)(func, paras_set)
                    opt()
                    func.reset()
                print('', end=')', flush=True)


    print('costs', round((time.time() - st) / 60, 2), 'minutes')
    with open("DataFiles/runningtime.txt", "a+") as text_file:
        date = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        runningtime = round((time.time() - st) / 60, 2)
        print("{} | {}: {} minutes.".format(date, optimizer_name, runningtime), file=text_file)





import numpy as np

class Init_Delta_Y:
    @staticmethod
    def your():
        pass
    @staticmethod
    def interval_type(M, n, interval): # pso, ba
        lb_y = interval[0]
        ub_y = interval[1]
        Y = np.random.uniform(lb_y, ub_y, (M, n))
        return Y
    @staticmethod
    def x_type(X): # mfo
        Y = X
        return Y




from .Init_X import Init_X
from .Init_Delta_X import Init_Delta_X
from .Init_Delta_Y import Init_Delta_Y
from .InitOpt_Delta_z import InitOpt_Delta_z
from .NatureOpt import NatureOpt
from .Opt_X import Opt_X
from .Opt_Delta_X import Opt_Delta_X
from .Opt_Delta_Y import Opt_Delta_Y
from .Selection import Selection
from .comparison import comparing


import numpy as np

class Selection:
    @staticmethod
    def your():
        pass

    @staticmethod
    def same_type(temp_X, temp_X_Fit): # pso, mfo, goa
        new_X = temp_X
        new_X_Fit = temp_X_Fit
        return new_X, new_X_Fit

    @staticmethod
    def elitism_type(temp_X, temp_X_Fit, old_X, old_X_Fit, elitism): # mbo
        # keep elitists in old_X
        sorted_old_X_inx = np.argsort(old_X_Fit)
        elitists_inx = sorted_old_X_inx[:elitism]
        elitists_x = old_X[elitists_inx]
        elitists_x_fit = old_X_Fit[elitists_inx]

        # replace keep worst x in temp_X by elitists
        new_X_inx = np.argsort(temp_X_Fit)
        new_X = temp_X[new_X_inx]
        new_X_Fit = temp_X_Fit[new_X_inx]
        for i in range(elitism):
            new_X[-(i + 1)] = elitists_x[i]
            new_X_Fit[-(i + 1)] = elitists_x_fit[i]
        return new_X, new_X_Fit

    @staticmethod
    def ba_type(temp_X, temp_X_Fit, old_X, old_X_Fit, z2):
        new_X = old_X.copy()
        new_X_Fit = old_X_Fit.copy()
        M = len(old_X)
        for i in range(M):
            if (np.random.rand() > z2) and (temp_X_Fit[i] < old_X_Fit[i]):
                new_X[i] = temp_X[i]
                new_X_Fit[i] = temp_X_Fit[i]
        return new_X, new_X_Fit

    @staticmethod
    def improve_type(temp_X, temp_X_Fit, old_X, old_X_Fit): # boa, csa
        M = len(temp_X)
        new_X = old_X.copy()
        new_X_Fit = old_X_Fit.copy()
        for i in range(M):
            if temp_X_Fit[i] < old_X_Fit[i]:
                new_X[i] = temp_X[i]
                new_X_Fit[i] = temp_X_Fit[i]
        return new_X, new_X_Fit



from UNIOA.NatureOpt import NatureOpt
# -------------------------------------------------------------------------------------------------
# This class implements MBO-Optimizer in the new structure.
# -------------------------------------------------------------------------------------------------

class MBO_Opt(NatureOpt):
    def __init__(self, func, hyperparams_set, budget_factor=1e4):
        super().__init__(func, budget_factor)
        self.M = hyperparams_set.get('popsize', 50)
        self.w1 = hyperparams_set.get('w1', 5 / 12)
        self.w2 = hyperparams_set.get('w2', 1.2)
        self.w3 = hyperparams_set.get('w3', 5 / 12)
        self.w4 = hyperparams_set.get('w4', 1)
        self.w5 = hyperparams_set.get('w5', 2)

    def __call__(self):
        t = 0
        X = self.Init_X.Init_X(M=self.M, n=self.n, lb_x=self.lb_x, ub_x=self.ub_x)  
        X_Fit = self.Evaluate_X(X=X)
        x_g, x_g_fit = self.Init_Delta_X.Global_best(new_X=X, new_X_Fit=X_Fit)
        z = self.InitOpt_Delta_z.mbo(t=t, w4=self.w4)

        # algorithm process
        while not self.stop:
            # OOptimize temp_X(t+1)
            temp_X = self.Opt_X.mbo(old_X=X, old_X_Fit=X_Fit, x_g=x_g, z=z, w1=self.w1, w2=self.w2, w3=self.w3, budget=self.budget)
            # Evaluate
            temp_X_Fit = self.Evaluate_X(X=temp_X)
            # Selection
            new_X, new_X_Fit = self.Selection.elitism_type(temp_X=temp_X, temp_X_Fit=temp_X_Fit, old_X=X, old_X_Fit=X_Fit,
                                                elitism=self.w5)
            # ----------------------------
            t = t + 1
            new_z = self.InitOpt_Delta_z.mbo(t=t, w4=self.w4)
            x_g, x_g_fit = self.Opt_Delta_X.Global_best(new_X=new_X, new_X_Fit=new_X_Fit, old_x_g=x_g, old_x_g_fit=x_g_fit)
            ########################################
            X = new_X
            X_Fit = new_X_Fit
            z = new_z




from .optimizer_running import optimizer_running
import sys,os


def comparing (Algs, problems, instances, dimensions, num_runs, paras_sets):
    '''
    Algs: a list of strings
    problems: a list of integers
    instances: a list of integers
    dimensions: a list of integers
    num_runs: one integer
    paras_set: dictionary
    '''
    problems = problems
    instances = instances
    dimensions = dimensions
    num_runs = num_runs

    for alg in Algs:
        optimizer_name = alg
        paras_set = paras_sets[alg]
        optimizer_running(problems, instances, dimensions, num_runs, paras_set, optimizer_name)





from UNIOA.NatureOpt import NatureOpt

# -------------------------------------------------------------------------------------------------
# This class implements Bat-Optimizer in the new structure.
# -------------------------------------------------------------------------------------------------
class BA_Opt(NatureOpt):
    def __init__(self, func ,hyperparams_set, budget_factor=1e4):
        super().__init__(func, budget_factor)
        self.M = hyperparams_set.get('popsize', 20)
        self.y_interval = hyperparams_set.get('y-interval', [0, 0])
        self.z1_0 = hyperparams_set.get('z1_0', 1)
        self.z2_0 = hyperparams_set.get('z2_0', 1)
        self.w1 = hyperparams_set.get('w1', 0.1)
        self.w2 = hyperparams_set.get('w2', 0.97)
        self.w3 = hyperparams_set.get('w3', 0.1)
        self.w4_interval = hyperparams_set.get('w4-interval', [0, 2])

    def __call__(self):
        t = 0
        X = self.Init_X.Init_X(M=self.M, n=self.n, lb_x=self.lb_x, ub_x=self.ub_x)
        X_Fit = self.Evaluate_X(X=X)  # Evaluate X
        Y = self.Init_Delta_Y.interval_type(M=self.M, n=self.n, interval=self.y_interval)
        x_g, x_g_fit = self.Init_Delta_X.Global_best(new_X=X, new_X_Fit=X_Fit)
        z1 = self.InitOpt_Delta_z.ba1(t=t, z1_0=self.z1_0, w1=self.w1)
        z2 = self.InitOpt_Delta_z.ba2(old_z2=self.z2_0, w2=self.w2)

        # Optimizing
        while not self.stop:
            # Aopt y(t+1)
            new_Y = self.Opt_Delta_Y.ba(old_Y=Y, old_X=X, old_x_g=x_g, w4_interval=self.w4_interval)
            # Oopt temp_x(t+1)
            temp_X = self.Opt_X.ba(old_X=X, new_Y=new_Y, old_x_g=x_g, old_z1=z1, old_z2=z2, w3=self.w3, lb_x=self.lb_x, ub_x=self.ub_x )
            # Evaluate
            temp_X_Fit = self.Evaluate_X(X=temp_X)
            # Selection
            new_X, new_X_Fit = self.Selection.ba_type(temp_X=temp_X, temp_X_Fit=temp_X_Fit, old_X=X, old_X_Fit=X_Fit,z2=z2)
            # ----------------------------
            t = t + 1
            new_x_g, new_x_g_fit = self.Opt_Delta_X.Global_best(new_X=new_X, new_X_Fit=new_X_Fit, old_x_g=x_g, old_x_g_fit=x_g_fit)
            new_z1 = self.InitOpt_Delta_z.ba1(t=t, z1_0=self.z1_0, w1=self.w1)
            new_z2 = self.InitOpt_Delta_z.ba2(old_z2=z2, w2=self.w2)
            ####################
            X = new_X
            X_Fit = new_X_Fit
            Y = new_Y
            x_g, x_g_fit = new_x_g, new_x_g_fit
            z1 = new_z1
            z2 = new_z2



 


import numpy as np

class Opt_Delta_Y:
    @staticmethod
    def your():
        pass
    @staticmethod
    def pso(old_X, old_Y, old_X_p, old_x_g, w1, w2, w3):
        new_Y = old_Y.copy()
        for i, old_y in enumerate(old_Y):
            new_Y[i] = w1 * old_y \
                       + np.random.uniform(0, w2) * (old_X_p[i] - old_X[i]) \
                       + np.random.uniform(0, w3) * (old_x_g - old_X[i])  # update equation.
            return new_Y
 
    @staticmethod
    def ba(old_Y, old_X, old_x_g, w4_interval):
        new_Y = old_Y.copy()
        for i in range(len(new_Y)):
            new_Y[i] = old_Y[i] + np.random.uniform(w4_interval[0], w4_interval[1]) * (old_X[i] - old_x_g)
        return new_Y

    @staticmethod
    def mfo(t, old_Y, fitness_function, old_X, old_X_Fit):
        old_Y_Fit = np.array([fitness_function(y) for y in old_Y])
        if t == 0:
            I = np.argsort(old_Y_Fit)
            new_Y = old_Y[I]
        else:
            Y_Fit_d = np.concatenate((old_Y_Fit, old_X_Fit))
            Y_d = np.concatenate((old_Y, old_X), axis=0)
            I_s_d = np.argsort(Y_Fit_d)
            new_Y_d = Y_d[I_s_d]
            new_Y = new_Y_d[:len(old_X)]

        # old_Y_Fit = np.array([fitness_function(y) for y in old_Y])
        # Y_Fit_d = np.concatenate((old_Y_Fit, old_X_Fit))
        # Y_d = np.concatenate((old_Y, old_X), axis=0)
        # I_s_d = np.argsort(Y_Fit_d)
        # new_Y_d = Y_d[I_s_d]
        # new_Y = new_Y_d[:len(old_X)]
        return new_Y

    @staticmethod
    def csa(old_Y, fitness_function, old_X, old_X_Fit):
       new_Y = old_Y.copy()
       M = len(old_X)
       for i in range(M):
           old_y_fit = fitness_function(old_Y[i])
           if old_X_Fit[i] < old_y_fit:
               new_Y[i] = old_X[i]
       return new_Y









from UNIOA.NatureOpt import NatureOpt

# -------------------------------------------------------------------------------------------------
# This class implements CSA-Optimizer in the new structure.
# -------------------------------------------------------------------------------------------------

class CSA_Opt(NatureOpt):
    def __init__(self, func ,hyperparams_set, budget_factor=1e4):
        super().__init__(func, budget_factor)
        self.M = hyperparams_set.get('popsize',50)
        self.w1 = hyperparams_set.get('w1', 0.1)
        self.w2 = hyperparams_set.get('w2', 2)


    
    def __call__(self):
        t = 0  # iteration counter
        X = self.Init_X.Init_X(M=self.M, n=self.n, lb_x=self.lb_x, ub_x=self.ub_x)
        X_Fit = self.Evaluate_X(X=X)  # Evaluate X
        Y = self.Init_Delta_Y.x_type(X=X)

        # Optimizing
        while not self.stop:
            # AOpt Y(t+1)
            new_Y = self.Opt_Delta_Y.csa(old_Y=Y, fitness_function=self.fitness_function, old_X=X, old_X_Fit=X_Fit)
            # OOpt temp_X(t+1)
            temp_X = self.Opt_X.csa(new_Y=new_Y, old_X=X, w1=self.w1, w2=self.w2, lb_x=self.lb_x, ub_x=self.ub_x)
            # Evaluate
            temp_X_Fit = self.Evaluate_X(X=temp_X)
            # Selection
            new_X, new_X_Fit = self.Selection.improve_type(temp_X=temp_X, temp_X_Fit=temp_X_Fit, old_X=X, old_X_Fit=X_Fit)
            # ----------------------------
            t = t + 1
            ##########################
            X = new_X
            X_Fit = new_X_Fit
            Y = new_Y




import numpy as np

class Init_X:
    @staticmethod
    def your():
        pass
    @staticmethod
    def Init_X(M, n, lb_x, ub_x):
        X = np.random.uniform(lb_x, ub_x, (M, n))
        return X




from setuptools import setup

setup(
    name='UNIOA',
    version='0.0.1',
    packages=['UNIOA', 'UNIOA.algs'],
    url='https://github.com/Huilin-Li/UNIOA.git',
    license='MIT',
    author='Huilin',
    author_email='lihuilin023@gmail.com',
    description='AutoNIOA'
)



from UNIOA.NatureOpt import NatureOpt
import numpy as np
# -------------------------------------------------------------------------------------------------
# This class implements PSO-Optimizer in the new structure.
# -------------------------------------------------------------------------------------------------

class PSO_Opt(NatureOpt):
    def __init__(self, func ,hyperparams_set, budget_factor=1e4):
        super().__init__(func, budget_factor)
        self.M = hyperparams_set.get('popsize', 25)
        self.y_interval = hyperparams_set.get('y-interval', [-1, 1])
        self.w1 = hyperparams_set.get('w1', 0.73)
        self.w2 = hyperparams_set.get('w2', 1.49)
        self.w3 = hyperparams_set.get('w3', 1.49)


    # algorithm process
    def __call__(self):
        t = 0
        X = self.Init_X.Init_X(M=self.M, n=self.n, lb_x=self.lb_x, ub_x=self.ub_x)
        X_Fit = self.Evaluate_X(X=X)
        Y = self.Init_Delta_Y.interval_type(M=self.M, n=self.n, interval=self.y_interval)
        X_p, X_p_Fit = self.Init_Delta_X.Personal_best(new_X=X, new_X_Fit=X_Fit)
        x_g, x_g_fit = self.Init_Delta_X.Global_best(new_X=X, new_X_Fit=X_Fit)
        while not self.stop:
            # Aoptimize y(t+1)
            new_Y = self.Opt_Delta_Y.pso(old_X=X, old_Y=Y, old_X_p=X_p, old_x_g=x_g, w1=self.w1, w2=self.w2, w3=self.w3)
            # Ooptimize x(t+1)
            temp_X = self.Opt_X.pso(old_X=X, new_Y=new_Y, lb_x=self.lb_x, ub_x=self.ub_x)
            # Evaluate
            temp_X_Fit = self.Evaluate_X(X=temp_X)
            # Selection
            new_X, new_X_Fit = self.Selection.same_type(temp_X=temp_X, temp_X_Fit=temp_X_Fit)
            #----------------------------
            t = t + 1
            new_X_p, new_X_p_Fit = self.Opt_Delta_X.Personal_best(new_X=new_X, new_X_Fit=new_X_Fit, old_X_p=X_p, old_X_p_Fit=X_p_Fit)
            new_x_g, new_x_g_fit = self.Opt_Delta_X.Global_best(new_X=new_X, new_X_Fit=new_X_Fit, old_x_g=x_g, old_x_g_fit=x_g_fit)
            ############################
            X = new_X
            #X_Fit = new_X_Fit
            Y = new_Y
            X_p, X_p_Fit = new_X_p, new_X_p_Fit
            x_g, x_g_fit = new_x_g, new_x_g_fit


from UNIOA.NatureOpt import NatureOpt

# -------------------------------------------------------------------------------------------------
# This class implements GOA-Optimizer in the new structure.
# -------------------------------------------------------------------------------------------------

class GOA_Opt(NatureOpt):
    def __init__(self, func, hyperparams_set, budget_factor=1e4):
        super().__init__(func, budget_factor)
        self.M = hyperparams_set.get('popsize', 100)
        self.z_interval = hyperparams_set.get('z-interval', [0.00004, 1])
        self.w1 = hyperparams_set.get('w1', 0.5)#=f
        self.w2 = hyperparams_set.get('w2', 1.5)#=l

    def __call__(self):
        t = 0
        X = self.Init_X.Init_X(M=self.M, n=self.n, lb_x=self.lb_x, ub_x=self.ub_x)
        X_Fit = self.Evaluate_X(X=X)
        x_g, x_g_fit = self.Init_Delta_X.Global_best(new_X=X, new_X_Fit=X_Fit)
        z = self.InitOpt_Delta_z.goa(t=t, z_interval=self.z_interval, T=self.budget)
        # Optimizing
        while not self.stop:
            # Oopt x(t+1)
            temp_X = self.Opt_X.goa(old_X=X, old_x_g = x_g, z=z, lb_x=self.lb_x, ub_x=self.ub_x, w1=self.w1, w2=self.w2)
            # Evaluate
            temp_X_Fit = self.Evaluate_X(X=temp_X)
            # Selection
            new_X, new_X_Fit = self.Selection.same_type(temp_X=temp_X, temp_X_Fit=temp_X_Fit)
            ########################
            t =  t + 1
            x_g, x_g_fit = self.Opt_Delta_X.Global_best(new_X=new_X, new_X_Fit=new_X_Fit, old_x_g=x_g, old_x_g_fit=x_g_fit)
            z = self.InitOpt_Delta_z.goa(t=t, z_interval=self.z_interval, T=self.budget)
            X = new_X




from .Init_X import Init_X
from .Init_Delta_X import Init_Delta_X
from .Init_Delta_Y import Init_Delta_Y
from .Opt_Delta_X import Opt_Delta_X
from .Opt_Delta_Y import Opt_Delta_Y
from .Opt_X import Opt_X
from .Selection import Selection
from .InitOpt_Delta_z import InitOpt_Delta_z
import numpy as np


class NatureOpt:
    def __init__(self, func, budget_factor = 1e4):
        # inherits objective problem, problem dimension, maximum iteration, problem boundary
        self.fitness_function = func
        self.n = func.meta_data.n_variables
        self.budget = budget_factor * self.n
        self.lb_x = self.fitness_function.constraint.lb[0]
        self.ub_x = self.fitness_function.constraint.ub[1]
        # inherits operations
        # compulsory components
        self.Init_X = Init_X
        self.Opt_X = Opt_X
        self.Selection = Selection
        # elective components
        self.Init_Delta_X = Init_Delta_X
        self.Init_Delta_Y = Init_Delta_Y
        self.InitOpt_Delta_z = InitOpt_Delta_z
        self.Opt_Delta_X = Opt_Delta_X
        self.Opt_Delta_Y = Opt_Delta_Y


    def Evaluate_X(self, X):
        fitness = [self.fitness_function(x) for x in X]
        return np.array(fitness)

    # def Personal_best(self, new_X, new_X_Fit, old_X_p = None, old_X_p_Fit = None):
    #     if old_X_p is None:
    #         new_X_p = new_X
    #         new_X_p_Fit = new_X_Fit
    #         return new_X_p, new_X_p_Fit
    #     else:
    #         new_X_p = old_X_p
    #         new_X_p_Fit = old_X_p_Fit
    #         for i, x in enumerate(new_X):
    #             if new_X_Fit[i] < old_X_p_Fit[i]:
    #                 new_X_p[i] = x
    #                 new_X_p_Fit[i] = new_X_Fit[i]
    #         return new_X_p, new_X_p_Fit
    #
    # def Global_best(self, new_X, new_X_Fit, old_x_g = None, old_x_g_fit = None):
    #     if old_x_g is None:
    #         best_index = np.argmin(new_X_Fit)
    #         x_g_fit = new_X_Fit[best_index]
    #         x_g = new_X[best_index]
    #         return x_g, x_g_fit
    #     else:
    #         xs = np.append([old_x_g],new_X, axis=0)
    #         fits = np.append([old_x_g_fit], new_X_Fit, axis=0)
    #         best_index = np.argmin(fits)
    #         x_g_fit = fits[best_index]
    #         x_g = xs[best_index]
    #         return x_g, x_g_fit

    @property
    def stop(self):
        return self.fitness_function.state.evaluations >= self.budget or self.fitness_function.state.current_best_internal.y < 1e-8
    